;;; PVS start-up file for Manip 1.2.  The definitions are evaluated
;;; when the Manip package is loaded.

(in-package :pvs)

(defparameter *manip-package-files*
  '("pregexp" "manip-utilities" "extended-expr"
    "syntax-matching" "manip-strategies"))

;;; Pathnames need to be discovered and constructed in a platform-
;;; and Lisp-independent way.  Avoid the use of "~" and other
;;; pathname shortcuts.

(defparameter *manip-home-dir*
  (loop for d in *pvs-library-path*
	for p = (merge-pathnames (make-pathname :name "Manip") d)
	when (file-exists-p p) return p))

(defparameter *manip-pathname-dir*
  (make-pathname :directory
		 (append (pathname-directory *manip-home-dir*) '("Manip"))))

(defparameter def-strat-names
  '(defrule defstrat defstep defhelper))

;;; Compile each package file when necessary.  First check all the
;;; top-level definitions (defxyz ...) to see if they might redefine
;;; anything in PVS or previously loaded packages.  Issue warnings
;;; if any symbol clashes are found.

(defun compile-package-file (lisp-file error-file)
  (flet ((rule-or-strategy (name)
	   (or (gethash name *rules*)
	       (gethash name *rulebase*)
	       (gethash name *steps*)))
	 (compile-file-with-log (file log-stream)
	   #+cmu
	   (compile-file file :error-output log-stream)
	   #-cmu
	   (let ((*standard-output* log-stream))
	     (compile-file file)) ))
    (let ((global-package-symbols nil)   ;; collect all def names found
	  (failed-global-symbols nil))   ;; collect those that clash
      (with-open-file (f lisp-file)
	(loop for form = (read f nil '<<<eof>>>)
	      if (eq form '<<<eof>>>) return nil
	      do (let ((op (car form)) (name (cadr form)))
		   (cond ((not (symbolp name)) nil)
			 ((memq name global-package-symbols) nil) ; seen already
;			 ((memq name failed-global-symbols) nil)
			 ((memq op def-strat-names)
			  (push name global-package-symbols)
			  (when (rule-or-strategy name)
			    ;; clash with predefined rule or strategy
			    (push name failed-global-symbols)))
			 ((not (equal (subseq (string op) 0 3) "def"))
			  nil)
			 ((or (boundp name) (fboundp name))
			  ;; clash with other kind of definition
			  (push name global-package-symbols)
			  (push name failed-global-symbols))
			 (t (push name global-package-symbols))))))

      ;; All detected symbol clashes are reported.  Some could be
      ;; harmless, though.  Let the user decide or report a problem.
      (when failed-global-symbols
	(warn "Manip: the symbols ~A are already defined."
	      failed-global-symbols)
	(pvs-emacs-eval
	 (format nil "(pvs-log-message \"ERR\" ~S)"
		 (format nil "Warning: Manip symbols ~A are already defined."
			 failed-global-symbols)))))
    (pvs-emacs-eval
     (format nil "(pvs-log-message \"MSG\" ~S)"
	     (format nil "Compiling file ~A..." lisp-file)))
    (compile-file-with-log lisp-file error-file)
    (pvs-emacs-eval
     (format nil "(pvs-log-message \"MSG\" ~S)"
	     (format nil "Finished compiling ~A." lisp-file)))))


;;; For each package file, check if its binary executable file is
;;; out of date, and if so, compile it.  Library loads are performed
;;; only on compiled files.

(flet ((form-pathname (name type)
         (merge-pathnames (make-pathname :name name :type type)
			  *manip-pathname-dir*)))
  (let* ((lisp-files (loop for f in *manip-package-files*
			   collect (form-pathname f "lisp")))
	 (exec-files (loop for f in *manip-package-files*
			   collect (form-pathname f *pvs-binary-type*)))
	 (out-of-date (loop for lisp-file in lisp-files
			    for exec-file in exec-files
			    unless (and (probe-file exec-file)
					(<= (file-write-date lisp-file)
					    (file-write-date exec-file)))
			      return t
			    finally (return nil))))

    ;; if any binary file is out-of-date, compile them all
    (when out-of-date
      (with-open-file (error-log (form-pathname "compile-messages" "log")
				 :direction :output :if-exists :append
				 :if-does-not-exist :create)
	(format error-log "~4%; =============== ~A =============== ~3%"
		(format nil "Compiling Manip on ~A"
			(date-string (get-universal-time))))
	(dolist (file lisp-files)
	  (compile-package-file file error-log))))

    ;; now load the binaries
    (dolist (file exec-files)
      ;; libload will generate log messages; no need to duplicate
      (libload file))))
